@@ -1,4 +1,5 @@
 
+#define PDL_CORE      /* For certain ifdefs */
 #include "pdl.h"      /* Data structure declarations */
 #include "pdlcore.h"  /* Core declarations */
 
@@ -47,6 +48,10 @@ int pdl_howbig (int datatype) {
     }
 }
 
+/* Note this is no longer called - will leave
+in for now until we are sure we are happy
+with the datatype conversion - KGB */
+
 int pdl_whichdatatype (double nv) {
 #define TESTTYPE(b,a) {a foo = nv; if(nv == foo) return b;}
 	TESTTYPE(PDL_B,PDL_Byte)
@@ -61,13 +66,13 @@ int pdl_whichdatatype (double nv) {
 
 /* Make a scratch data existence for a pdl */
 
-void pdl_makescratchhash(pdl *ret,double data) {
+void pdl_makescratchhash(pdl *ret,double data, int datatype) {
 	HV *hash;
 	SV *dat; PDL_Long fake[1];
 
 	 /* Compress to smallest available type. This may have strange
 	    results sometimes :( */
-	ret->datatype = pdl_whichdatatype(data);
+	ret->datatype = datatype;
 	ret->data = pdl_malloc(pdl_howbig(ret->datatype)); /* Wasteful */
 
        dat = newSVpv(ret->data,pdl_howbig(ret->datatype));
@@ -106,17 +111,26 @@ pdl* SvPDLV ( SV* sv ) {
    SV *sv2;
 
    if ( !SvROK(sv) ) {   /* Coerce scalar */
-	SV *dat;
-	double data;
+      SV *dat;
+      double data;
+      int datatype;
+	
        ret = pdl_new();  /* Scratch pdl */
 
 /*       ret->sv = (void*) sv; !! */
 
 /* Scratch hash for the pdl :( - slow but safest. */
 
-	data = SvNV(sv);
+       
+       /* Figure datatype to use */
 
-	pdl_makescratchhash(ret,data);
+       if ( ((SvIOK(sv) && !SvNOK(sv))) || !SvNIOK(sv) )  /* Int */
+          datatype = PDL_L;
+       else /* Double */
+          datatype = PDL_D;
+       
+       data = SvNV(sv);
+       pdl_makescratchhash(ret,data,datatype);
 
        return ret;
    }
@@ -127,9 +141,36 @@ pdl* SvPDLV ( SV* sv ) {
 	if(svp == NULL) {
 		croak("Hash given as a pdl - but not {PDL} key!");
 	}
-   	sv = *svp;
+	
+	/* This is the magic hook which checks to see if {PDL} 
+	is a code ref, and if so executes it. It should
+	return a standard piddle. This allows
+	all kinds of funky objects to be derived from PDL,
+	and allow normal PDL functions to still work so long
+	as the {PDL} code returns a standard piddle on
+	demand - KGB */
+	
+	if (SvROK(*svp) && SvTYPE(SvRV(*svp)) == SVt_PVCV) {
+	   dSP;
+	   int count;
+	   ENTER ;
+	   SAVETMPS;
+	   PUSHMARK(sp) ;
+	   count = perl_call_sv(*svp, G_SCALAR|G_NOARGS);
+	   SPAGAIN ;
+	   if (count != 1)
+              croak("Execution of PDL structure failed to return one value\n") ;
+
+	   sv=newSVsv(POPs);
+	   PUTBACK ;
+	   FREETMPS ;
+	   LEAVE ;	
+	}
+	else {
+   	   sv = *svp;
+	}
    }
-       
+          
    if (SvTYPE(SvRV(sv)) != SVt_PVMG)
       croak("Error - argument is not a recognised data structure"); 
 
@@ -280,5 +321,128 @@ void* pdl_malloc ( int nbytes ) {
    return (void *) SvPV(work, na);
 }
 
+/*********** Stuff for barfing *************/
 
+/* Version of perl's mess_alloc - we need this copy
+here because it is defined static in util.c! */
 
+static SV *
+pdl_mess_alloc()
+{
+    SV *sv;
+    XPVMG *any;
+
+    /* Create as PVMG now, to avoid any upgrading later */
+    New(905, sv, 1, SV);
+    Newz(905, any, 1, XPVMG);
+    SvFLAGS(sv) = SVt_PVMG;
+    SvANY(sv) = (void*)any;
+    SvREFCNT(sv) = 1 << 30; /* practically infinite */
+    return sv;
+}
+
+/* Version of perl's mess() constructor which doesn't
+do the automatic appending of stuff when "\n" not
+seen at the end of the string - for use by pdl_barf() */
+
+char *
+pdl_mess(pat, args)
+    const char *pat;
+    va_list *args;
+{
+    SV *sv;
+
+    if (!mess_sv)
+	mess_sv = pdl_mess_alloc();
+    sv = mess_sv;
+    sv_vsetpvfn(sv, pat, strlen(pat), args, Null(SV**), 0, Null(bool*));
+    {
+	/* call the PDL message enhancing routine */
+	ENTER;
+	LEAVE;
+        {
+	    dSP;
+	    SV *msg;
+
+	    ENTER;
+
+	    PUSHMARK(sp);
+	    XPUSHs(sv);
+	    PUTBACK;
+	    perl_call_pv("PDL::Core::barf_msg", G_SCALAR);
+            sv = POPs;
+	    LEAVE;
+	}
+    }
+    return SvPVX(sv);
+}
+
+/*
+   pdl_barf - warning routine to be called when PDL routine
+   croak 
+   
+   Note: we call back in to Perl as error reporting is
+   not time critical, and this gives us lots of flexibility.
+   
+   code stolen from croak() in util.c in Perl distribution
+   
+*/
+   
+#ifdef I_STDARG
+void
+pdl_barf(const char* pat, ...)
+#else
+/*VARARGS0*/
+void
+pdl_barf(pat, va_alist)
+    char *pat;
+    va_dcl
+#endif
+{
+    va_list args;
+    char *message;
+    HV *stash;
+    GV *gv;
+    CV *cv;
+
+#ifdef I_STDARG
+    va_start(args, pat);
+#else
+    va_start(args);
+#endif
+    message = pdl_mess(pat, &args);
+    va_end(args);
+     
+    if (diehook) {
+	/* sv_2cv might call croak() */
+	SV *olddiehook = diehook;
+	ENTER;
+	SAVESPTR(diehook);
+	diehook = Nullsv;
+	cv = sv_2cv(olddiehook, &stash, &gv, 0);
+	LEAVE;
+	if (cv && !CvDEPTH(cv) && (CvROOT(cv) || CvXSUB(cv))) {
+	    dSP;
+	    SV *msg;
+
+	    ENTER;
+	    msg = newSVpv(message, 0);
+	    SvREADONLY_on(msg);
+	    SAVEFREESV(msg);
+
+	    PUSHMARK(sp);
+	    XPUSHs(msg);
+	    PUTBACK;
+	    perl_call_sv((SV*)cv, G_DISCARD);
+
+	    LEAVE;
+	}
+    }
+    if (in_eval) {
+	restartop = die_where(message);
+	JMPENV_JUMP(3);
+    }
+    PerlIO_puts(PerlIO_stderr(),message);
+    (void)PerlIO_flush(PerlIO_stderr());
+    my_failure_exit();
+}
