@@ -11,12 +11,16 @@ sub get_pdls {my($this) = @_; return $this->{ParNames};}
 # Do the appropriate substitutions in the code.
 sub new { my($type,$code,$parnames,$parobjs,$indobjs,$generictypes,
 	    $extrageneric,$havethreading) = @_;
+	my $sizeprivs = {};
 	my($this) = bless {
 		IndObjs => $indobjs,
 		ParNames => $parnames,
 		ParObjs => $parobjs,
+		Gencurtype => [], # stack to hold GenType in generic loops
 	},$type;
-	$_ = "{$code}";
+	$_ = "{".
+	   (join '',map {$_->get_incregisters();} (values %{$this->{ParObjs}})).
+	    "$code}";
 # First, separate the code into an array of C fragments (strings),
 # variable references (strings starting with $) and
 # loops (array references, 1. item = variable.
@@ -27,9 +31,9 @@ sub new { my($type,$code,$parnames,$parobjs,$indobjs,$generictypes,
 	while($_) {
 # Parse next statement
 		s/^(.*?) # First, some noise is allowed. This may be bad.
-		   ( \$[a-zA-Z_]+\([^)]*\)   # $a(...): access
-		    |\bloop\([^)]+\)\s*%{   # loop(..) %{
-		    |\btypes\([^)]+\)\s*%{  # types(..) %{
+		   ( \$[a-zA-Z_]+\s*\([^)]*\)   # $a(...): access
+		    |\bloop\s*\([^)]+\)\s*%{   # loop(..) %{
+		    |\btypes\s*\([^)]+\)\s*%{  # types(..) %{
 		    |\bthreadloop\s*%{        # threadloop %{
 		    |%}                     # %}
 		    |$)//xs
@@ -40,11 +44,13 @@ sub new { my($type,$code,$parnames,$parobjs,$indobjs,$generictypes,
 		push @{$stack[-1]},$1;
 # Then, our control.
 		if($control) {
-			if($control =~ /^loop\(([^)]+)\)\s*%{/) {
-				my $ob = new PDL::PP::Loop([split ',',$1]);
+			if($control =~ /^loop\s*\(([^)]+)\)\s*%{/) {
+				my $ob = new PDL::PP::Loop([split ',',$1],
+					$sizeprivs,$this);
+				print "SIZEPRIVSXX: $sizeprivs,",(join ',',%$sizeprivs),"\n" if $::PP_VERBOSE;
 				push @{$stack[-1]},$ob;
 				push @stack,$ob;
-			} elsif($control =~ /^types\(([^)]+)\)\s*%{/) {
+			} elsif($control =~ /^types\s*\(([^)]+)\)\s*%{/) {
 				my $ob = new PDL::PP::Types($1);
 				push @{$stack[-1]},$ob;
 				push @stack,$ob;
@@ -53,7 +59,7 @@ sub new { my($type,$code,$parnames,$parobjs,$indobjs,$generictypes,
 				push @{$stack[-1]},$ob;
 				push @stack,$ob;
 				$threadloops ++;
-			} elsif($control =~ /^\$[a-zA-Z_]+\([^)]*\)/) {
+			} elsif($control =~ /^\$[a-zA-Z_]+\s*\([^)]*\)/) {
 				push @{$stack[-1]},new PDL::PP::Access($control,$this);
 			} elsif($control =~ /^%}/) {
 				pop @stack;
@@ -61,9 +67,10 @@ sub new { my($type,$code,$parnames,$parobjs,$indobjs,$generictypes,
 				confess("Invalid control: $control\n");
 			}
 		} else {
-			print("No \$2!\n");
+			print("No \$2!\n") if $::PP_VERBOSE;
 		}
 	}
+	print "SIZEPRIVSX: ",(join ',',%$sizeprivs),"\n" if $::PP_VERBOSE;
 # Now, if there is no explicit threadlooping in the code,
 # enclose everything into it.
 	if(!$threadloops && $havethreading) {
@@ -93,7 +100,10 @@ sub new { my($type,$code,$parnames,$parobjs,$indobjs,$generictypes,
 		push @$coderef,$nc;
 	}
 # Then, in this form, put it together what we want the code to actually do.
-	$this->{Code} = $coderef->get_str($this,[]);
+	print "SIZEPRIVS: ",(join ',',%$sizeprivs),"\n" if $::PP_VERBOSE;
+	$this->{Code} = "{".(join '',values %$sizeprivs).
+	  $coderef->get_str($this,[])
+	  ."}";
 	$this->{Code};
 }
 
@@ -152,8 +162,16 @@ sub get_str {my ($this,$parent,$context) = @_;
 package PDL::PP::Loop;
 @PDL::PP::Loop::ISA = PDL::PP::Block;
 
-sub new { my($type,$args) = @_;
-	bless [$args],$type;
+sub new { my($type,$args,$sizeprivs,$parent) = @_;
+	my $this = bless [$args],$type;
+	for(@{$this->[0]}) {
+		print "SIZP $sizeprivs, $_\n" if $::PP_VERBOSE;
+		my $i = $parent->make_loopind($_);
+		$sizeprivs->{$i->[0]} = 
+		  "register int __$i->[0]_size = \$PRIV(__$i->[0]_size);\n";
+		print "SP :",(join ',',%$sizeprivs),"\n" if $::PP_VERBOSE;
+	}
+	return $this;
 }
 
 sub myoffs { return 1; }
@@ -161,8 +179,9 @@ sub myprelude { my($this,$parent,$context) = @_;
 	my $text = ""; my $i;
 	push @$context, map {
 		$i = $parent->make_loopind($_);
-		$text .= "{/* Open $_ */ long $_; 
-			for($_=0; $_<\$PRIV(__$i->[0]_size); $_++) {";
+# Used to be $PRIV(.._size) but now we have it in a register.
+		$text .= "{/* Open $_ */ register int $_; 
+			for($_=0; $_<(__$i->[0]_size); $_++) {";
 		$i;
 	} @{$this->[0]};
 	return $text;
@@ -189,6 +208,7 @@ sub myoffs {4}
 
 
 sub myprelude { my($this,$parent,$context) = @_;
+	push @{$parent->{Gencurtype}},'PDL_undef'; # so that $GENERIC can get at it
 	"/* Start generic loop */\n".
 	(join '',map{
 		"#undef THISIS$this->[1]_$_\n#define THISIS$this->[1]_$_(a)\n"
@@ -199,6 +219,7 @@ sub myprelude { my($this,$parent,$context) = @_;
 sub myitem { my($this,$parent,$nth) = @_;
 	my $item = $this->[0]->[$nth];
 	if(!$item) {return "";}
+	$parent->{Gencurtype}->[-1] = $item->[1];
 	"\t} break; case $item->[0]: {\n".
 	(join '',map {
 		"#undef THISIS$this->[1]_$_\n#define THISIS$this->[1]_$_(a)\n";
@@ -210,8 +231,9 @@ sub myitem { my($this,$parent,$nth) = @_;
 }
 
 sub mypostlude { my($this,$parent,$context) = @_;
+	pop @{$parent->{Gencurtype}};  # and clean up the Gentype stack
 	"\tbreak;}
-	default:croak(\"PP INTERNAL ERROR! PLEASE MAKE A BUG REPORT\\n\");}\n";
+	default:barf(\"PP INTERNAL ERROR! PLEASE MAKE A BUG REPORT\\n\");}\n";
 }
 
 
@@ -235,7 +257,8 @@ sub myprelude {my($this,$parent,$context) = @_;
  my $no;
  my ($ord,$pdls) = $parent->get_pdls();
 '	/* THREADLOOPBEGIN */
- PDL->startthreadloop(&($PRIV(__thread)));
+ if(PDL->startthreadloop(&($PRIV(__thread)),$PRIV(vtable)->readdata,
+ 	__privtrans))) return;
    do { 
  '.(join '',map {"${_}_datap += \$PRIV(__thread).offs[".(0+$no++)."];\n"}
  		@$ord).'
@@ -265,15 +288,26 @@ use Carp;
 sub new { my($type) = @_; bless [],$type; }
 sub myoffs { return 0; }
 sub myprelude {my($this,$parent,$context) = @_;
- my $no;
+ my $no; my $no2=-1; my $no3=-1;
  my ($ord,$pdls) = $parent->get_pdls();
 '	/* THREADLOOPBEGIN */
- PDL->startthreadloop(&($PRIV(__thread)));
-   do { int __tind1=0,__tind2=0;  int __tnpdls = $PRIV(__thread).npdls;
- '.(join '',map {"${_}_datap += \$PRIV(__thread).offs[".(0+$no++)."];\n"}
+ if(PDL->startthreadloop(&($PRIV(__thread)),$PRIV(vtable)->readdata,
+ 	__tr)) return;
+   do { register int __tind1=0,__tind2=0;  
+        register int __tnpdls = $PRIV(__thread).npdls;
+	register int __tdims1 = $PRIV(__thread.dims[1]);
+	register int __tdims0 = $PRIV(__thread.dims[0]);
+	register int *__offsp = PDL->get_threadoffsp(&$PRIV(__thread));
+ '.(join '',map {$no2++;
+            "register int __tinc0_".($no2)." = \$PRIV(__thread).incs[$no2];"}
+	     @$ord).
+   (join '',map {$no3++;
+            "register int __tinc1_".($no3)." = \$PRIV(__thread).incs[__tnpdls+$no3];"}
+	     @$ord).
+   (join '',map {"${_}_datap += __offsp[".(0+$no++)."];\n"}
  		@$ord).'
-	for(__tind2=0; __tind2<$PRIV(__thread.dims[1]); __tind2++) {
-	 for(__tind1=0; __tind1<$PRIV(__thread.dims[0]); __tind1++) {
+	for(__tind2=0; __tind2<__tdims1 ; __tind2++) {
+	 for(__tind1=0; __tind1<__tdims0 ; __tind1++) {
 	  /* This is the tightest threadloop. Make sure inside is optimal. */
 ';
 }
@@ -283,16 +317,16 @@ sub mypostlude {my($this,$parent,$context) = @_;
  my $no; my $no0; my $no1; my $no2; my $no3; my $no4; my $no5;
  my ($ord,$pdls) = $parent->get_pdls();
 '	/* THREADLOOPEND */
-	 '.(join '',map {"${_}_datap += \$PRIV(__thread).incs[".(0+$no0++)."];\n"}
+	 '.(join '',map {"${_}_datap += __tinc0_".(0+$no0++).";\n"}
  		@$ord).'
 	 } '
-	 .(join '',map {"${_}_datap += \$PRIV(__thread).incs[__tnpdls+".(0+$no1++)."]
-	     			     - \$PRIV(__thread).incs[".(0+$no2++)."] *
-				       \$PRIV(__thread).dims[0];\n"}
+	 .(join '',map {"${_}_datap += __tinc1_".(0+$no1++)."
+	     			     - __tinc0_".(0+$no2++)." *
+				       __tdims0;\n"}
  		@$ord).'
 	} '.
-     (join '',map {"${_}_datap -= \$PRIV(__thread).incs[__tnpdls+".(0+$no3++)."] *
-     				  \$PRIV(__thread).dims[1];"}
+     (join '',map {"${_}_datap -= __tinc1_".(0+$no3++)." *
+     				  __tdims1;"}
  		@$ord).'
  '.(join '',map {"${_}_datap -= \$PRIV(__thread).offs[".(0+$no++)."];\n"}
  		@$ord).'
@@ -331,12 +365,14 @@ package PDL::PP::Access;
 use Carp;
 
 sub new { my($type,$str,$parent) = @_;
-	$str =~ /^\$([a-zA-Z_]+)\(([^)]*)\)/ or
+	$str =~ /^\$([a-zA-Z_]+)\s*\(([^)]*)\)/ or
 		confess ("Access wrong: $access\n");
 	my($pdl,$inds) = ($1,$2);
 	if($pdl =~ /^T/) {new PDL::PP::MacroAccess($pdl,$inds);} 
 	elsif($pdl =~ /^P$/) {new PDL::PP::PointerAccess($pdl,$inds);}
 	elsif($pdl =~ /^PP$/) {new PDL::PP::PhysPointerAccess($pdl,$inds);}
+        elsif($pdl =~ /^SIZE$/) {new PDL::PP::SizeAccess($pdl,$inds);}
+        elsif($pdl =~ /^GENERIC$/) {new PDL::PP::GentypeAccess($pdl,$inds);}
 	elsif(!defined $parent->{ParObjs}{$pdl}) {new PDL::PP::OtherAccess($pdl,$inds);}
 	else {
 		bless [$pdl,$inds],$type;
@@ -368,8 +404,10 @@ use Carp;
 sub new { my($type,$pdl,$inds) = @_; bless [$inds],$type; }
 
 sub get_str {my($this,$parent,$context) = @_;
-	$parent->{ParObjs}{$this->[0]}->do_pointeraccess()
-	 if defined($parent->{ParObjs}{$this->[0]});
+	croak ("can't access undefined pdl ".$this->[0])
+	  unless defined($parent->{ParObjs}{$this->[0]});
+	$parent->{ParObjs}{$this->[0]}->{FlagPaccess} = 1;
+	$parent->{ParObjs}{$this->[0]}->do_pointeraccess();
 }
 
 
@@ -402,12 +440,56 @@ sub get_str {my($this,$parent,$context) = @_;
 	my @lst = split ',',$inds;
 	my @ilst = split '',$1;
 	if($#lst != $#ilst) {confess("Macroaccess: different nos of args $pdl $inds\n");}
-	return join ' ',map {
-		"THISIS_$ilst[$_]($lst[$_])"
-	} (0..$#lst) ;
+	croak "generic type access outside a generic loop"
+	  unless defined $parent->{Gencurtype}->[-1];
+	my $type = $parent->{Gencurtype}->[-1];
+	croak "unknown Type" unless $type =~ /^PDL_([BSULFD])/;
+	$type = $1;
+	for (0..$#lst) {
+	  return "$lst[$_]" if $ilst[$_] =~ /$type/;
+	}
+	# return join ' ',map {
+	#	"THISIS_$ilst[$_]($lst[$_])"
+	# } (0..$#lst) ;
 }
 
 
+###########################
+# 
+# Encapsulate a SizeAccess
+
+package PDL::PP::SizeAccess;
+use Carp;
+
+sub new { my($type,$pdl,$inds) = @_; bless [$inds],$type; }
+
+sub get_str {my($this,$parent,$context) = @_;
+	croak "can't get SIZE of undefined dimension $this->[0]"
+	  unless defined($parent->{IndObjs}{$this->[0]});
+	$parent->{IndObjs}{$this->[0]}->get_size();
+}
+
+###########################
+# 
+# Encapsulate a GentypeAccess
+
+package PDL::PP::GentypeAccess;
+use Carp;
+
+sub new { my($type,$pdl,$inds) = @_; bless [$inds],$type; }
+
+sub get_str {my($this,$parent,$context) = @_;
+	     croak "generic type access outside a generic loop"
+	       unless defined $parent->{Gencurtype}->[-1];
+	     my $type = $parent->{Gencurtype}->[-1];
+	     if ($this->[0]) {
+	       croak "not a defined name"
+		 unless defined($parent->{ParObjs}{$this->[0]});
+	       $type = $parent->{ParObjs}{$this->[0]}->ctype($type);
+	     }
+	     return $type;
+}
+
 ########################
 #
 # Type coercion
@@ -464,4 +546,3 @@ sub xxx_get_generictypes { my($this) = @_;
 		[F,"PDL_Float",$PDL_F],
 		[D,"PDL_Double",$PDL_D])];
 }
-
