@@ -2,6 +2,13 @@
 
 PDL::Gaussian -- Gaussian distributions.
 
+=head1 SYNOPSIS
+
+	$a = new PDL::Gaussian([3],[5]);
+	$a->set_covariance(...)
+
+	# XXXX Must improve this interface
+
 =head1 DESCRIPTION
 
 This package provides a set of standard routines to handle 
@@ -94,23 +101,26 @@ of the given arguments for reasonability.
 
 =head1 AUTHOR
 
-Copyright (C) 1996 Tuomas J. Lukka (lukka@fas.harvard.edu)
+Copyright (C) 1996 Tuomas J. Lukka (lukka@fas.harvard.edu) Reproducing 
+documentation from the pdl distribution in any way that does not include a
+statement telling who the original authors are is forbidden.  Reproducing and/or
+distributing the documentation  in any  form that  alters the text is
+forbidden.
 
 =cut
 
 package PDL::Gaussian;
 
+use PDL::Core '';
 use PDL::Slatec;
 use PDL::Primitive;
 use PDL::Slice;
 use PDL::Lib::LLSQRout;
 
-use Carp;
-
 sub new {
 	my($type,$ndims,$nfuncs) = @_;
 	if($#{$ndims} != 0) {
-		confess("PDL::Gaussian can only have one dimension dimensionality\n");
+		barf("PDL::Gaussian can only have one dimension dimensionality\n");
 	}
 	my $ndims1 = ($#{$ndims==0} ? $ndims : [1]);
 	bless {
